import logging
import sys
import threading

from ..frontends.constrained_frontend import ConstrainedFrontend

l = logging.getLogger(__name__)


class SMTLibScriptDumperMixin:
    def get_smtlib_script_satisfiability(self, extra_constraints=(), extra_variables=()):
        """
        Return an smt-lib script that check the satisfiability of the current constraints

        :return string: smt-lib script
        """
        try:
            e_csts = self._solver_backend.convert_list(extra_constraints + tuple(self.constraints))
            e_variables = self._solver_backend.convert_list(extra_variables)

            variables, csts = self._solver_backend._get_all_vars_and_constraints(e_c=e_csts, e_v=e_variables)
            return self._solver_backend._get_satisfiability_smt_script(csts, variables)
        except BackendError as e:
            raise ClaripyFrontendError("Backend error during smtlib script generation") from e

    # def merge(self, others, merge_conditions, common_ancestor=None):
    #     return self._solver_backend.__class__.__name__ == 'BackendZ3', ConstrainedFrontend.merge(
    #         self, others, merge_conditions, common_ancestor=common_ancestor
    #     )[1]


from ..errors import BackendError, ClaripyFrontendError



import unittest
import claripy
from claripy.backends.backend_smtlib_solvers.cvc4_popen import SolverBackendCVC4
import common_backend_smt_solver


class SmtLibSolverTest_CVC4(common_backend_smt_solver.SmtLibSolverTestBase):
    @common_backend_smt_solver.if_installed
    def get_solver(self):
        backend = SolverBackendCVC4(daggify=True)
        solver = claripy.SolverStrings(backend=backend, timeout=300000)
        return solver


if __name__ == "__main__":
    suite = unittest.TestLoader().loadTestsFromTestCase(SmtLibSolverTest_CVC4)
    unittest.TextTestRunner(verbosity=2).run(suite)



# hits = 0
# misses = 0
# ejects = 0


class CompositedCacheMixin:
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._merged_solvers = {}

    def _blank_copy(self, c):
        super()._blank_copy(c)
        c._merged_solvers = {}

    def _copy(self, c):
        super()._copy(c)
        c._merged_solvers = dict(self._merged_solvers)

    def __setstate__(self, base_state):
        super().__setstate__(base_state)
        self._merged_solvers = {}

    #
    # Cache stuff
    #

    def _remove_cached(self, names):
        # global ejects

        for k in list(self._merged_solvers.keys()):
            if k & names:
                # ejects += 1
                self._merged_solvers.pop(k)

    def _solver_for_names(self, names):
        # global hits, misses

        n = frozenset(names)
        try:
            r = self._merged_solvers[frozenset(n)]
            # hits += 1
            return r
        except KeyError:
            # misses += 1
            s = super()._solver_for_names(names)
            self._merged_solvers[n] = s
            return s

    def downsize(self):
        super().downsize()
        self._merged_solvers = {}

    def _store_child(self, s, **kwargs):
        self._remove_cached(s.variables)
        return super()._store_child(s, **kwargs)



import logging

l = logging.getLogger("claripy.ops")

#
# AST creation
#


def AbstractLocation(*args, **kwargs):  # pylint:disable=no-self-use
    aloc = vsa.AbstractLocation(*args, **kwargs)
    return aloc


#
# Some operations
#


#
# sigh
#

# pylint:disable=wildcard-import,unused-wildcard-import
from .ast.base import *
from .ast.bv import *
from .ast.fp import *
from .ast.bool import *
from .ast.strings import *
from . import vsa

VS = ValueSet
